"""The Keys lock device implementation"""
import logging

from . import Action, TheKeysDevice

OPENED = "Door open"
CLOSED = "Door closed"
JAMMED = "Door jammed"
UNKNOWN = ""

_LOGGER = logging.getLogger(__name__)


class TheKeysLock(TheKeysDevice):
    """The Keys lock device implementation"""

    def __init__(self, name: str, host: str, identifier: str, share_code: str) -> None:
        super().__init__(
            name=name, host=host, identifier=identifier, share_code=share_code
        )
        self.locker_status = UNKNOWN
        self.__retrieve_lock_status()

    def lock(self) -> bool:
        """Lock this lock"""
        response = self.action(Action.CLOSE)
        if response["status"] == "ok":
            self.locker_status = CLOSED
        else:
            self.locker_status = JAMMED

        return response["status"] == "ok"

    def unlock(self) -> bool:
        """Unlock this lock"""
        response = self.action(Action.OPEN)
        if response["status"] == "ok":
            self.locker_status = OPENED
        else:
            self.locker_status = JAMMED

        return response["status"] == "ok"

    def update(self) -> None:
        """Update the lock status"""
        self.__retrieve_lock_status()

    @property
    def is_unlocked(self) -> bool:
        """Is this lock unlocked"""
        return self.locker_status == OPENED

    @property
    def is_locked(self) -> bool:
        """Is this lock locked"""
        return self.locker_status == CLOSED

    @property
    def is_jammed(self) -> bool:
        """Is this lock jammed"""
        return self.locker_status == JAMMED

    def __retrieve_lock_status(self) -> None:
        try:
            self.locker_status = self.action(Action.STATUS)["status"]
        except ConnectionError as error:
            _LOGGER.error(error)



"""TheKeys devices"""
import base64
import hmac
import time
from enum import Enum
from typing import Any

import requests


class Action(Enum):
    """All available actions"""
    OPEN = "open"
    CLOSE = "close"
    STATUS = "locker_status"

    def __str__(self):
        return self.value


class TheKeysDevice:
    """Base class for all TheKeys devices"""

    def __init__(self, name: str, host: str, identifier: str, share_code: str) -> None:
        self.name = name
        self.host = host
        self.identifier = identifier
        self.share_code = share_code

    def update(self) -> None:
        """TODO"""

    def action(self, action: Action) -> Any:
        """Execute action on the device"""
        timestamp = str(int(time.time()))
        hash = base64.b64encode(hmac.new(self.share_code.encode("ascii"), timestamp.encode("ascii"), "sha256").digest())

        response = requests.post(
            f"{self.host}/{action}",
            data={"hash": hash, "identifier": self.identifier, "ts": timestamp},
        )
        return response.json()



"""
Python library to handle the keys api
"""
from typing import Any, List, Optional
from dataclasses import dataclass, field
from dataclasses_json import LetterCase, config, dataclass_json
import requests

from .devices import TheKeysDevice
from .devices.lock import TheKeysLock

BASE_URL = "https://api.the-keys.fr"


@dataclass_json
@dataclass
class DatabaseActionDate:
    """Data class for timestamp"""
    date: str
    timezone_type: int
    timezone: str


@dataclass_json
@dataclass
class LockUser:
    """Data class for lock user"""
    username: str
    firstname: str
    lastname: str


@dataclass_json
@dataclass
class AccessoryInfo:
    """Data class for accessory info"""
    last_seen: str
    ip: str


@dataclass_json
@dataclass
class Accessory:
    """Data class for accessory"""
    id: int
    id_accessoire: str
    nom: str
    type: int
    version: int
    type_version: int
    created_at: DatabaseActionDate
    updated_at: DatabaseActionDate
    public_key: str
    info: AccessoryInfo


@dataclass_json
@dataclass
class LockAccessory:
    """Data class for lock accessory"""
    id: int
    id_accessoire: str
    nom: str
    type: int
    configuration: Any


@dataclass_json
@dataclass
class LockAccessoryHolder:
    """Data class for lock accessory"""
    id: int
    accessoire: LockAccessory
    info: Optional[str]


@dataclass_json
@dataclass
class LockProduct:
    """Data class for lock product"""
    id: int
    nom: str
    version: int
    version_beta: int = field(metadata=config(letter_case=LetterCase.CAMEL))


@dataclass_json
@dataclass
class Lock:
    """Data class for lock"""
    id: str
    id_serrure: str
    code: str
    code_serrure: str
    etat: str
    couleur: Optional[str]
    nom: str
    qrcode: str
    serrure_droite: bool
    main_libre: bool
    longitude: int
    latitude: int
    radius: int
    timezone: str
    max_speed: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    latch_delay: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    assisted_actions: bool = field(metadata=config(letter_case=LetterCase.CAMEL))
    unlock_only: bool = field(metadata=config(letter_case=LetterCase.CAMEL))
    description: Optional[str]
    log_sequence: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    public_key: str
    message: str
    utilisateur: LockUser
    version: int
    version_cible: int
    beta: int
    battery: int
    battery_date: DatabaseActionDate
    accessoires: List[LockAccessoryHolder]
    produit: LockProduct


@dataclass_json
@dataclass
class User:
    """Data class for user"""
    id: str
    type: str
    roles: List[str]
    firstname: str
    lastname: str
    locale: str
    username: str
    email: str
    created_at: DatabaseActionDate
    updated_at: DatabaseActionDate
    notification_token: str
    notification_enabled: bool
    serrures: List[Lock]
    tel: str


@dataclass_json
@dataclass
class ShareRole:
    """Data class for share role"""
    id: int
    description: str


@dataclass_json
@dataclass
class ShareUser:
    """Data class for share user"""
    username: str
    prenom: str
    nom: str
    email: str
    telephone: str


@dataclass_json
@dataclass
class AbstractShare:
    """Data class for abstract share"""
    id: int
    nom: str
    actif: bool
    date_debut: Optional[str]
    date_fin: Optional[str]
    heure_debut: Optional[str]
    heure_fin: Optional[str]
    description: Optional[str]
    notification_enabled: bool
    horaires: List[Any]


@dataclass_json
@dataclass
class UserShare:
    """Data class for user share"""
    role: ShareRole
    utilisateur: ShareUser
    remote_key_sharing_id: int = field(metadata=config(letter_case=LetterCase.CAMEL))


@dataclass_json
@dataclass
class ShareAccessory:
    """Data class for share accessory"""
    id: int
    id_accessoire: str
    nom: str
    type: int
    version: int
    type_version: int
    configuration: List[Any]


@dataclass_json
@dataclass
class AccessoryShare:
    """Data class for accessory share"""
    iddesc: str
    accessoire: ShareAccessory
    code: str


@dataclass_json
@dataclass
class Share:
    """Data class for share"""
    partages_utilisateur: List[UserShare]
    partages_accessoire: List[AccessoryShare]


class TheKeyApi:
    """TheKeysApi class"""

    def __init__(self, username, password) -> None:
        self.username = username
        self.password = password
        self.access_token = None
        self.__authenticate()

    @property
    def authenticated(self):
        """Get the token"""
        return not self.access_token is None

    def find_user_by_username(self, username: str) -> User:
        """Return user matching the passed username"""
        return User.from_dict(self.__http_get(f"utilisateur/get/{username}")["data"])

    def find_lock_by_id(self, id: int) -> Lock:
        """Return lock matching the passed id"""
        return Lock.from_dict(self.__http_get(f"serrure/get/{id}")["data"])

    def find_accessory_by_id(self, id: int) -> Accessory:
        """Return accessory matching the passed id"""
        return Accessory.from_dict(self.__http_get(f"accessoire/get/{id}")["data"])

    def find_share_by_lock_id(self, lock_id: int) -> Share:
        """Return share matching the passed lock_id"""
        return Share.from_dict(self.__http_get(f"partage/all/serrure/{lock_id}")["data"])

    def get_devices(self) -> List[TheKeysDevice]:
        """Return all devices"""
        devices = []
        user = self.find_user_by_username(self.username)
        for lock in user.serrures:
            shares = self.find_share_by_lock_id(lock.id)

            accessory = next(
                (x for x in lock.accessoires if x.accessoire.type == 1),
                None,
            )
            gateway = self.find_accessory_by_id(accessory.accessoire.id)

            share = next(
                x
                for x in shares.partages_accessoire
                if x.accessoire.id == gateway.id
            )

            devices.append(
                TheKeysLock(
                    name=lock.nom,
                    host=f"http://{gateway.info.ip}",
                    identifier=lock.id_serrure,
                    share_code=share.code,
                )
            )

        return devices

    def __http_get(self, url: str):
        if not self.authenticated:
            self.__authenticate()

        response = requests.get(
            f"{BASE_URL}/fr/api/v2/{url}",
            headers={"Authorization": f"Bearer {self.access_token}"},
        )
        if response.status_code != 200:
            raise RuntimeError()

        return response.json()

    def __authenticate(self):
        response = requests.post(
            f"{BASE_URL}/api/login_check",
            data={"_username": self.username, "_password": self.password},
        )
        if response.status_code == 200:
            json = response.json()
            self.access_token = json["access_token"]
            self.expires_in = json["expires_in"]
        else:
            raise RuntimeError("Access denied")

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if exception_type is not None:
            print(exception_value)

        return True



from setuptools import find_packages, setup

setup(
    name='the_keyspy',
    packages=find_packages(include=('the_keyspy*',)),
    version='0.0.9',
    description='The Keys Api',
    author='Kevin Bonnoron',
    author_email='kevin.bonnoron@gmail.com',
    url='https://github.com/KevinBonnoron',
    download_url='https://github.com/KevinBonnoron/the_keyspy/archive/refs/tags/v0.0.9.tar.gz',
    license='MIT',
    install_requires=['dataclasses_json', 'requests'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==6.2.4'],
    test_suite='tests',
)
